@@ -33,7 +33,7 @@ module PuppetDBExtensions
 
     platform_version =
       get_option_value(options[:platform_version],
-                       [:puppet5, :puppet6, :puppet7],
+                       [:puppet5, :puppet6, :puppet7, :puppet8, :"puppet8-nightly"],
                        "Puppet Platform Repo Version",
                        "PLATFORM_VERSION",
                        :puppet5)
@@ -197,12 +197,7 @@ module PuppetDBExtensions
   def get_os_family(host)
     on(host, "which yum", :silent => true)
     if result.exit_code == 0
-      on(host, "ls /etc/fedora-release", :silent => true)
-      if result.exit_code == 2
-        :redhat
-      else
-        :fedora
-      end
+      :redhat
     else
       :debian
     end
@@ -289,6 +284,10 @@ module PuppetDBExtensions
     return test_config[:os_families].has_key? 'ubuntu2004-64-1'
   end
 
+  def is_jammy()
+    return test_config[:os_families].has_key? 'ubuntu2204-64-1'
+  end
+
   def is_buster()
     return test_config[:os_families].has_key? 'debian10-64-1'
   end
@@ -297,16 +296,15 @@ module PuppetDBExtensions
     return test_config[:os_families].has_key? 'debian11-64-1'
   end
 
-  def is_el6()
-    return test_config[:os_families].has_key?('redhat6-64-1') ||
-           test_config[:os_families].has_key?('centos6-64-1')
-  end
-
   def is_el8()
     return test_config[:os_families].has_key?('redhat8-64-1') ||
            test_config[:os_families].has_key?('centos8-64-1')
   end
 
+  def is_el9()
+    return test_config[:os_families].has_key?('redhat9-64-1')
+  end
+
   def is_rhel7fips
     return test_config[:os_families].has_key? 'redhatfips7-64-1'
   end
@@ -314,21 +312,35 @@ module PuppetDBExtensions
   # This specifies which puppet platform repo to initialize in order
   # to find packages.
   #
+  # For install tests it should pick the nightly repo so we are testing the
+  # latest dev builds of agent and server
+  #
+  # For upgrade_latest tests it should be the non-nightly version. Tests are
+  # run with the shipped package because nightly packages are pruned eventually.
+  #
   # For upgrade_oldest tests it must match the result of oldest_supported
   #
   # TODO: when testing X.0.0-SNAPSHOT one of install/upgrade_latest tests
   # will fail because those two tests will need to use different repos
-  def puppet_repo_version
-    case test_config[:install_mode]
-    when :install, :upgrade_latest
-      test_config[:platform_version]
-    when :upgrade_oldest
-      # Debian 11 only has builds starting in the 7 series
-      if is_bullseye
-        :puppet7
+  def puppet_repo_version(platform_version, install_type, nightly)
+    # When adding a new platform version, only puppet8-nightly will be available
+    # for a time. In order to generate packages we will have to run with the nightly
+    # platform version (in addition to our own nightly boolean). This will prune the -nightly
+    # and allow us to function normally.
+    sanitized_version = "#{platform_version}".delete_suffix("-nightly").to_sym
+
+    case install_type
+    when :install
+      if nightly
+        :"#{sanitized_version}-nightly"
       else
-        :puppet6
+        sanitized_version
       end
+    when :upgrade_latest
+      # always use the non-nightly version to install released packages
+      sanitized_version
+    when :upgrade_oldest
+      :puppet7
     end
   end
 
@@ -336,19 +348,13 @@ module PuppetDBExtensions
   # supported version of PuppetDB. Its version must be available in the
   # platform version returned by puppet_repo_version above
   def oldest_supported
-    # account for bionic/rhel8 not having build before certian versions
-    if is_bullseye
+    # account for bionic/rhel8 not having build before certain versions
+    if is_el9 || is_jammy
+      '7.14.0'
+    elsif is_bullseye
       '7.9.0'
-    elsif is_el8
-      '6.0.3'
-    elsif is_rhel7fips
-      '6.4.0'
-    elsif is_buster
-      '6.7.0'
-    elsif is_focal
-      '6.12.0'
     else
-      '6.0.0'
+      '7.3.1'
     end
   end
 
@@ -404,27 +410,18 @@ module PuppetDBExtensions
     ## These 'platform' values come from the acceptance config files, so
     ## we're relying entirely on naming conventions here.  Would be nicer
     ## to do this using lsb_release or something, but...
-    if host['platform'].include?('el-5')
-      "#{version}.el5"
-    elsif host['platform'].include?('el-6')
-      "#{version}.el6"
-    elsif host['platform'].include?('el-7')
+    if host['platform'].include?('el-7')
       "#{version}.el7"
     elsif host['platform'].include?('el-8')
       "#{version}.el8"
-    elsif host['platform'].include?('fedora')
-      version_tag = host['platform'].match(/^fedora-(\d+)/)[1]
-      "#{version}.fc#{version_tag}"
-    elsif host['platform'].include?('ubuntu-16.04')
-      "#{version}xenial"
+    elsif host['platform'].include?('el-9')
+      "#{version}.el9"
     elsif host['platform'].include?('ubuntu-18.04')
       "#{version}bionic"
     elsif host['platform'].include?('ubuntu-20.04')
       "#{version}focal"
-    elsif host['platform'].include?('debian-8')
-      "#{version}jessie"
-    elsif host['platform'].include?('debian-9')
-      "#{version}stretch"
+    elsif host['platform'].include?('ubuntu-22.04')
+      "#{version}jammy"
     elsif host['platform'].include?('debian-10')
       "#{version}buster"
     elsif host['platform'].include?('debian-11')
@@ -434,6 +431,12 @@ module PuppetDBExtensions
     end
   end
 
+  def install_puppetdb_module(hosts, puppet_platform)
+    # While we sort out a new puppetlabs-puppetdb module release, point to a branch that allows us to take the latest puppetlabs-postgresql module
+    on(hosts, 'curl -L https://github.com/puppetlabs/puppetlabs-puppetdb/archive/refs/heads/bump-postgres.tar.gz --output /tmp/puppetlabs-puppetdb')
+    on(hosts, 'puppet module install /tmp/puppetlabs-puppetdb')
+  end
+
   def install_puppetdb(host, version=nil)
     manifest = <<-EOS
     class { 'puppetdb::globals':
@@ -459,7 +462,7 @@ module PuppetDBExtensions
           when :debian
             result = on host, "dpkg-query --showformat \"\\${Version}\" --show puppetdb"
             result.stdout.strip
-          when :redhat, :fedora
+          when :redhat
             result = on host, "rpm -q puppetdb --queryformat \"%{VERSION}-%{RELEASE}\""
             result.stdout.strip
           else
@@ -518,7 +521,7 @@ module PuppetDBExtensions
 
   def enable_https_apt_sources(host)
     manifest = <<-EOS
-      if ($facts['osfamily'] == 'Debian') {
+      if ($facts['os']['family'] == 'Debian') {
         package { 'apt-transport-https': ensure => installed }
       }
     EOS
@@ -527,12 +530,15 @@ module PuppetDBExtensions
   end
 
   def postgres_manifest
+    # bionic is EOL, so its pgdg repo has been remove
+    manage_package_repo = ! is_bionic
+
     manifest = <<-EOS
       # create the puppetdb database
       class { '::puppetdb::database::postgresql':
       listen_addresses            => 'localhost',
-      manage_package_repo         =>  true,
-      postgres_version            => '11',
+      manage_package_repo         => #{manage_package_repo},
+      postgres_version            => '14',
       database_name               => 'puppetdb',
       database_username           => 'puppetdb',
       database_password           => 'puppetdb',
@@ -743,14 +749,6 @@ EOS
     install_postgres(host)
   end
 
-  ############################################################################
-  # NOTE: This code was merged into beaker, however it does not work as desired.
-  #   We need to get the version in beaker working as expected and then we can
-  #   remove this version.
-  #
-  #   Temp copy of Justins new Puppet Master Methods
-  ############################################################################
-
   # Restore puppet.conf from backup, if puppet.conf.bak exists.
   #
   # @api private
@@ -763,10 +761,6 @@ EOS
              "fi"
   end
 
-  ##############################################################################
-  # END_OF Temp Copy of Justins new Puppet Master Methods
-  ##############################################################################
-
   def parse_json_with_error(input)
     begin
       facts = JSON.parse(input)
@@ -856,125 +850,18 @@ EOS
   # End Object diff functions
   ##############################################################################
 
-  def initialize_repo_on_host(host, os, nightly, puppet_platform = :puppet6)
-    case os
-    when :debian
-
-      # For openjdk8
-      if host['platform'].version == '8'
-        create_remote_file(host,
-                          "/etc/apt/sources.list.d/jessie-backports.list",
-                          "deb https://artifactory.delivery.puppetlabs.net/artifactory/debian_archive__remote/ jessie-backports main")
-      end
-
-      if options[:type] == 'aio' then
-        if nightly && puppet_platform == :puppet6
-          ## puppet6 repos
-          on host, "curl -O https://nightlies.puppet.com/apt/puppet6-nightly-release-$(lsb_release -sc).deb"
-          on host, "dpkg -i puppet6-nightly-release-$(lsb_release -sc).deb"
-
-        elsif nightly && puppet_platform == :puppet5
-          on host, "curl -O https://nightlies.puppet.com/apt/puppet5-nightly-release-$(lsb_release -sc).deb"
-          on host, "dpkg -i puppet5-nightly-release-$(lsb_release -sc).deb"
-
-        elsif nightly && puppet_platform == :puppet7
-          on host, "curl -O https://nightlies.puppet.com/apt/puppet7-nightly-release-$(lsb_release -sc).deb"
-          on host, "dpkg -i puppet7-nightly-release-$(lsb_release -sc).deb"
-
-        else
-          on host, "curl -O https://apt.puppet.com/puppet7-release-$(lsb_release -sc).deb"
-          on host, "dpkg -i puppet7-release-$(lsb_release -sc).deb"
-        end
-      else
-        on host, "curl -O https://apt.puppet.com/puppet6-release-$(lsb_release -sc).deb"
-        on host, "dpkg -i puppet6-release-$(lsb_release -sc).deb"
-      end
-      on host, "apt-get update"
-      on host, "apt-get install debian-archive-keyring"
-    when :redhat
-      if options[:type] == 'aio' then
-        /^(el|centos)-(\d+)-(.+)$/.match(host.platform)
-        variant = ($1 == 'centos') ? 'el' : $1
-        version = $2
-        _arch = $3
-
-        if nightly && puppet_platform == :puppet6
-          ## puppet6 repos
-          on host, "curl -O https://nightlies.puppet.com/yum/puppet6-nightly-release-#{variant}-#{version}.noarch.rpm"
-          on host, "rpm -i puppet6-nightly-release-#{variant}-#{version}.noarch.rpm"
-
-        elsif nightly && puppet_platform == :puppet5
-          ## puppet6 repos
-          on host, "curl -O https://nightlies.puppet.com/yum/puppet5-nightly-release-#{variant}-#{version}.noarch.rpm"
-          on host, "rpm -i puppet5-nightly-release-#{variant}-#{version}.noarch.rpm"
-
-        elsif nightly && puppet_platform == :puppet7
-          on host, "curl -O https://nightlies.puppet.com/yum/puppet7-nightly-release-#{variant}-#{version}.noarch.rpm"
-          on host, "rpm -i puppet7-nightly-release-#{variant}-#{version}.noarch.rpm"
-
-        else
-          on host, "curl -O https://yum.puppet.com/puppet7-release-#{variant}-#{version}.noarch.rpm"
-          on host, "rpm -i puppet7-release-#{variant}-#{version}.noarch.rpm"
-        end
-      else
-        on host, "yum clean all -y"
-        on host, "yum upgrade -y"
-        create_remote_file host, '/etc/yum.repos.d/puppetlabs-dependencies.repo', <<-REPO.gsub(' '*8, '')
-  [puppetlabs-dependencies]
-  name=Puppet Labs Dependencies - $basearch
-  baseurl=https://yum.puppetlabs.com/el/$releasever/dependencies/$basearch
-  gpgkey=https://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
-  enabled=1
-  gpgcheck=1
-        REPO
-
-        create_remote_file host, '/etc/yum.repos.d/puppetlabs-products.repo', <<-REPO.gsub(' '*8, '')
-  [puppetlabs-products]
-  name=Puppet Labs Products - $basearch
-  baseurl=https://yum.puppetlabs.com/el/$releasever/products/$basearch
-  gpgkey=https://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
-  enabled=1
-  gpgcheck=1
-        REPO
-
-        create_remote_file host, '/etc/yum.repos.d/epel.repo', <<-REPO
-  [epel]
-  name=Extra Packages for Enterprise Linux $releasever - $basearch
-  baseurl=https://download.fedoraproject.org/pub/epel/$releasever/$basearch
-  mirrorlist=https://mirrors.fedoraproject.org/metalink?repo=epel-$releasever&arch=$basearch
-  failovermethod=priority
-  enabled=1
-  gpgcheck=0
-        REPO
-      end
-    when :fedora
-      create_remote_file host, '/etc/yum.repos.d/puppetlabs-dependencies.repo', <<-REPO.gsub(' '*8, '')
-  [puppetlabs-dependencies]
-  name=Puppet Labs Dependencies - $basearch
-  baseurl=https://yum.puppetlabs.com/fedora/f$releasever/dependencies/$basearch
-  gpgkey=https://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
-  enabled=1
-  gpgcheck=1
-      REPO
-
-      create_remote_file host, '/etc/yum.repos.d/puppetlabs-products.repo', <<-REPO.gsub(' '*8, '')
-  [puppetlabs-products]
-  name=Puppet Labs Products - $basearch
-  baseurl=https://yum.puppetlabs.com/fedora/f$releasever/products/$basearch
-  gpgkey=https://yum.puppetlabs.com/RPM-GPG-KEY-puppetlabs
-  enabled=1
-  gpgcheck=1
-      REPO
-    else
-      raise ArgumentError, "Unsupported OS '#{os}'"
-    end
-  end
-
   def install_puppet_from_package(puppet_collection = "puppet6")
     hosts.each do |host|
       install_puppet_agent_on(host, {:puppet_collection => puppet_collection})
       on( host, puppet('resource', 'host', 'updates.puppetlabs.com', 'ensure=present', "ip=127.0.0.1") )
-      install_package(host, 'puppetserver')
+
+      # Take new configuration file on debian to prevent tests from hanging
+      # when this is used to upgrade puppet
+      if get_os_family(database) == :debian
+        database.install_package('puppetserver', '-o Dpkg::Options::="--force-confnew"')
+      else
+        install_package(database, 'puppetserver')
+      end
     end
   end
 
@@ -1045,7 +932,7 @@ EOS
       os = os_families[host.name]
 
       case os
-      when :redhat, :fedora
+      when :redhat
         install_pacakge(host, 'ruby')
         install_pacakge(host, 'git-core')
       when :debian
@@ -1090,11 +977,6 @@ EOS
   end
 
   def install_puppet(puppet_collection = "puppet6")
-    hosts.each do |host|
-      if host['platform'].variant == 'debian' && host['platform'].version == '8'
-        install_package(host, "openjdk-8-jre-headless")
-      end
-    end
     # If our :install_type is :pe then the harness has already installed puppet.
     case test_config[:install_type]
     when :package
@@ -1113,51 +995,6 @@ EOS
     install_puppet_conf
   end
 
-  def create_remote_site_pp(host, manifest)
-    testdir = host.tmpdir("remote-site-pp")
-    manifest_file = "#{testdir}/environments/production/manifests/site.pp"
-    apply_manifest_on(host, <<-PP)
-    File {
-      ensure => directory,
-      mode => "0750",
-      owner => #{master.puppet['user']},
-      group => #{master.puppet['group']},
-    }
-
-    file {
-      '#{testdir}':;
-      '#{testdir}/environments':;
-      '#{testdir}/environments/production':;
-      '#{testdir}/environments/production/manifests':;
-      '#{testdir}/environments/production/modules':;
-    }
-PP
-    create_remote_file(host, manifest_file, manifest)
-    remote_path = "#{testdir}/environments"
-    on host, "chmod -R +rX #{testdir}"
-    on host, "chown -R #{master.puppet['user']}:#{master.puppet['user']} #{testdir}"
-    remote_path
-  end
-
-  def run_agents_with_new_site_pp(host, manifest, env_vars = {}, extra_cli_args = "")
-
-    manifest_path = create_remote_site_pp(host, manifest)
-    with_puppet_running_on host, {
-      'master' => {
-        'storeconfigs' => 'true',
-        'storeconfigs_backend' => 'puppetdb',
-        'autosign' => 'true',
-      },
-      'main' => {
-        'environmentpath' => manifest_path,
-      }} do
-      #only some of the opts work on puppet_agent, acceptable exit codes does not
-      agents.each{ |agent| on agent, puppet_agent("--test --server #{host} #{extra_cli_args}",
-                                                  { 'ENV' => env_vars }), :acceptable_exit_codes => [0,2] }
-
-    end
-  end
-
   def databases
     extend Beaker::DSL::Roles
     hosts_as(:database).sort_by {|db| db.to_str}
