@@ -3,12 +3,13 @@
             [clojure.java.io :as io]
             [clojure.tools.logging :as log]
             [clojure.set :refer [rename-keys]]
-            [murphy :refer [with-open!]]
+            [murphy :refer [with-open! try!]]
             [puppetlabs.i18n.core :refer [trs tru]]
             [puppetlabs.puppetdb.query.paging :as paging]
             [puppetlabs.puppetdb.cheshire :as json]
             [puppetlabs.puppetdb.http :as http]
             [puppetlabs.puppetdb.jdbc :as jdbc]
+            [puppetlabs.puppetdb.query-eng.default-reports :as dr]
             [puppetlabs.puppetdb.query.aggregate-event-counts :as aggregate-event-counts]
             [puppetlabs.puppetdb.query.edges :as edges]
             [puppetlabs.puppetdb.query.events :as events]
@@ -17,10 +18,11 @@
             [puppetlabs.puppetdb.query.fact-contents :as fact-contents]
             [puppetlabs.puppetdb.query.resources :as resources]
             [puppetlabs.puppetdb.query.catalog-inputs :as inputs]
-            [puppetlabs.puppetdb.query-eng.default-reports :as dr]
+            [puppetlabs.puppetdb.query.monitor :as qmon]
             [puppetlabs.puppetdb.scf.storage-utils :as sutils]
             [puppetlabs.puppetdb.schema :as pls]
-            [puppetlabs.puppetdb.utils :as utils :refer [with-log-mdc]]
+            [puppetlabs.puppetdb.utils :as utils
+             :refer [with-log-mdc time-limited-seq]]
             [puppetlabs.puppetdb.utils.string-formatter :as formatter]
             [puppetlabs.puppetdb.query-eng.engine :as eng]
             [ring.util.io :as rio]
@@ -29,9 +31,28 @@
    (clojure.lang ExceptionInfo)
    (com.fasterxml.jackson.core JsonParseException)
    (java.io IOException InputStream)
+   (java.nio.channels SelectionKey)
+   (java.sql SQLException)
    (org.postgresql.util PGobject PSQLException)
    (org.joda.time Period)))
 
+(defn query-terminated-msg [id origin]
+  (if origin
+    (trs "PDBQuery:{0}: from {1} was terminated in Postgres" id (pr-str origin))
+    (trs "PDBQuery:{0}: was terminated in Postgres" id)))
+
+(defn query-terminated [id origin]
+  ;; Note: the exception message is currently logged directly in some cases.
+  (ex-info (query-terminated-msg id origin)
+           {:kind :puppetlabs.puppetdb.query/terminated :id id :origin origin}))
+
+(defn query-timeout [id origin]
+  ;; Note: the exception message is currently logged directly in some cases.
+  (ex-info (if origin
+             (trs "PDBQuery:{0}: from {1} exceeded timeout" id (pr-str origin))
+             (trs "PDBQuery:{0}: exceeded timeout" id))
+           {:kind :puppetlabs.puppetdb.query/timeout :id id :origin origin}))
+
 (def entity-fn-idx
   (atom
    {:aggregate-event-counts {:munge aggregate-event-counts/munge-result-rows
@@ -58,7 +79,7 @@
     :catalog-input-contents {:munge (constantly identity)
                              :rec eng/catalog-input-contents-query}
     :catalog-inputs {:munge inputs/munge-catalog-inputs
-                             :rec eng/catalog-inputs-query}
+                     :rec eng/catalog-inputs-query}
     :nodes {:munge (constantly identity)
             :rec eng/nodes-query}
     ;; Not a real entity name of course -- requested by query parameter.
@@ -183,10 +204,6 @@
 
         (munge-fn version url-prefix)))))
 
-(def use-preferred-streaming-method?
-  (->> (or (System/getenv "PDB_USE_DEPRECATED_QUERY_STREAMING_METHOD") "")
-       (re-matches #"yes|true|1") seq not))
-
 (def query-context-schema
   {:scf-read-db s/Any
    :url-prefix String
@@ -195,17 +212,20 @@
    (s/optional-key :warn-experimental) Boolean
    (s/optional-key :pretty-print) (s/maybe Boolean)
    (s/optional-key :log-queries) Boolean
-   (s/optional-key :query-id) s/Str})
+   (s/optional-key :query-id) s/Str
+   (s/optional-key :query-deadline-ns) s/Num
+   (s/optional-key :query-monitor) s/Any
+   (s/optional-key :query-monitor-id) SelectionKey
+   (s/optional-key :puppetlabs.puppetdb.config/test) {s/Any s/Any}})
 
 (pls/defn-validated stream-query-result
-  "Given a query, and database connection, return a Ring response with the query
-   results."
+  "Given a query, and database connection, call row-fn on the
+  resulting (munged) row sequence."
   ([version query options context]
    ;; We default to doall because tests need this for the most part
    (stream-query-result version query options context doall))
   ([version :- s/Keyword
-    query
-    options
+    query options
     context :- query-context-schema
     row-fn]
    ;; For now, generate the ids here; perhaps later, higher up
@@ -218,24 +238,48 @@
        (let [{:keys [scf-read-db url-prefix warn-experimental log-queries]
               :or {warn-experimental true}} context
              {:keys [remaining-query entity]} (eng/parse-query-context query warn-experimental)
-             munge-fn (get-munge-fn entity version options url-prefix)]
-
+             munge-fn (get-munge-fn entity version options url-prefix)
+             throw-timeout #(throw (query-timeout query-id origin))
+             {:keys [query-deadline-ns query-monitor query-monitor-id]} context]
          (when log-queries
            ;; Log origin and AST of incoming query
            (log/infof "PDBQuery:%s:%s"
                       query-id (-> (sorted-map :origin origin :ast query)
                                    json/generate-string)))
-
-         (let [f #(let [{:keys [results-query]}
-                        (query->sql remaining-query entity version
-                                    (merge options
-                                           (select-keys context [:node-purge-ttl :add-agent-report-filter]))
-                                    (select-keys context [:log-queries :query-id]))]
-                    (jdbc/call-with-array-converted-query-rows results-query
-                                                               (comp row-fn munge-fn)))]
-           (if use-preferred-streaming-method?
-             (jdbc/with-db-connection scf-read-db (jdbc/with-db-transaction [] (f)))
-             (jdbc/with-transacted-connection scf-read-db (f)))))))))
+         (jdbc/with-db-connection scf-read-db
+           (when query-monitor-id
+             (qmon/register-pg-pid query-monitor query-monitor-id
+                                   (jdbc/current-pid)))
+           (try!
+             (jdbc/with-db-transaction []
+               (try
+                 (jdbc/update-local-timeouts query-deadline-ns 1)
+                 (let [{:keys [results-query]}
+                       (query->sql remaining-query entity version
+                                   (merge options
+                                          (select-keys context [:node-purge-ttl :add-agent-report-filter]))
+                                   (select-keys context [:log-queries :query-id]))]
+                   (jdbc/call-with-array-converted-query-rows
+                    results-query
+                    (cond-> (comp row-fn munge-fn)
+                      query-deadline-ns
+                      (comp #(time-limited-seq % query-deadline-ns throw-timeout)))))
+                 (catch ExceptionInfo ex
+                   ;; Recast pg_terminate_backend errors since most(?)
+                   ;; will be intentional via the query monitor, and
+                   ;; we can't reliably distinguish others.
+                   ;; cf. wrap-with-exception-handling.
+                   (if (= (jdbc/sql-state :admin-shutdown)
+                          (some-> (ex-data ex) :handling .getSQLState))
+                     (throw (query-terminated query-id origin))
+                     (throw ex)))
+                 (catch SQLException ex
+                   (if (jdbc/local-timeout-ex? ex)
+                     (throw-timeout)
+                     (throw ex)))))
+             (finally
+               (when query-monitor-id
+                 (qmon/forget query-monitor query-monitor-id))))))))))
 
 ;; Do we still need this, i.e. do we need the pass-through, and the
 ;; strict selectivity in the caller below?
@@ -302,16 +346,24 @@
       (reset [] (.reset stream))
       (skip [n] (.skip stream n)))))
 
+;; Could this just be a key in a query/context specific map?
+
+;; Strictly for testing (seconds) - only works if all the threads
+;; involved are related via binding conveyance.  Otherwise, we'll need
+;; a redef.
+(def ^:dynamic diagnostic-inter-row-sleep nil)
+
 (defn- body-stream
   "Returns a map whose :stream is an InputStream that produces (via a
   future) the results of the query as JSON, and whose :status is a
   promise to which nil or {:count n} will be delivered after the first
-  row is retrieved, or to which {:error exception} will be delivered
-  if an exception happens before then.  An exception thrown by the
-  future after that point will produce an exception from the next call
-  to the InputStream read or close methods."
+  row is retrieved (if any), or to which {:error exception} will be
+  delivered if an exception happens before then.  An exception thrown
+  by the future after that point will produce an exception from the
+  next call to the InputStream read or close methods."
   [db query entity version query-options munge-fn
-   {:keys [pretty-print query-id] :as context}]
+   {:keys [pretty-print query-id query-deadline-ns query-monitor query-monitor-id]
+    :as context}]
   ;; Client disconnects present as generic IOExceptions from the
   ;; output writer (via stream-json), and we just log them at debug
   ;; level.  For now, after the first row, there's nothing we can do
@@ -326,163 +378,132 @@
   ;;
   ;; produce-streaming-body blocks until status is delivered, so
   ;; ensure it always is.
-  (let [status (promise)
+
+  (let [origin (:origin query-options)
+        ;; This is used for cases where the consumer (jetty thread) is
+        ;; presumed "gone", e.g. if the attempt to write json to it
+        ;; throws an IO exception.  In that case, we don't want to
+        ;; hand it an exception to log (presumably it already did
+        ;; something), we just want to handle it ourselves and quit.
+        ;; This must only be used after we've delivered the status,
+        ;; since it (intentionally) doesn't deliver.
         quiet-exit (Exception. "private singleton escape exception escaped")
-        stream (generated-stream
-                ;; Runs in a future
-                (fn [out]
+        throw-timeout #(throw (query-timeout query-id origin))
+        status (promise)
+        ;; The time-limited-seq should check for timeout before and
+        ;; after realizing each row, which should allow us to time out
+        ;; between the db pull and network push.  If we ever want to
+        ;; handle db-side errors independently, we'll need to add a
+        ;; pull-side try/catch (e.g., I think, the (seq s) in
+        ;; time-delmited-seq) or to rewrite this as an explicit loop,
+        ;; or...
+        stream-rows (fn stream-rows [rows out status-after-query-start]
+                      (let [timeout-seq #(time-limited-seq % query-deadline-ns throw-timeout)
+                            rows (cond->> rows
+                                   diagnostic-inter-row-sleep (map #(dissoc % :pg_sleep))
+                                   true munge-fn
+                                   query-deadline-ns timeout-seq)]
+                        (when-not (instance? PGobject rows)
+                          (first rows))
+                        (deliver status status-after-query-start)
+                        (try
+                          (http/stream-json rows out pretty-print)
+                          (catch IOException ex
+                            (log/debug ex (trs "Unable to stream response: {0}"
+                                               (.getMessage ex)))
+                            (throw quiet-exit)))))]
+
+    ;; Summarize all pg_terminate_backend errors in here since most(?)
+    ;; will be produced by the query monitor, and we can't reliably
+    ;; distinguish others.  cf. wrap-with-exception-handling which
+    ;; handles the throws.
+    {:status status
+     :stream (generated-stream
+              ;; Runs in a future
+              (fn serialize-query-response [out]
+                (try
                   (with-log-mdc ["pdb-query-id" query-id
-                                 "pdb-query-origin" (:origin query-options)]
+                                 "pdb-query-origin" origin]
                     (with-open! [out (io/writer out :encoding "UTF-8")]
                       (try
                         (jdbc/with-db-connection db
-                          (jdbc/with-db-transaction []
-                            (let [{:keys [results-query count-query]}
-                                  (query->sql query entity version query-options context)
-                                  st (when count-query
-                                       {:count (jdbc/get-result-count count-query)})
-                                  stream-row (fn [row]
-                                               (let [r (munge-fn row)]
-                                                 (when-not (instance? PGobject r)
-                                                   (first r))
-                                                 (when-not (realized? status)
-                                                   (deliver status st))
-                                                 (try
-                                                   (http/stream-json r out pretty-print)
-                                                   (catch IOException ex
-                                                     (log/debug ex (trs "Unable to stream response: {0}"
-                                                                        (.getMessage ex)))
-                                                     (throw quiet-exit)))))]
-                              (jdbc/call-with-array-converted-query-rows results-query
-                                                                         stream-row)
-                              (when-not (realized? status)
-                                (deliver status st)))))
-                        (catch Exception ex
-                          ;; If it's an exit, we've already handled it.
-                          (when-not (identical? quiet-exit ex)
-                            (if (realized? status)
-                              (throw ex)
-                              (deliver status {:error ex}))))
-                        (catch Throwable ex
-                          (if (realized? status)
-                            (do
-                              (log/error ex (trs "Query streaming failed: {0} {1}"
-                                                 query query-options))
-                              (throw ex))
-                            (deliver status {:error ex}))))))))]
-    {:status status
-     :stream stream}))
+                          (when query-monitor-id
+                            (qmon/register-pg-pid query-monitor query-monitor-id
+                                                  (jdbc/current-pid)))
+                          (try!
+                            (jdbc/with-db-transaction []
+                              (jdbc/update-local-timeouts query-deadline-ns 1)
+                              (let [{:keys [results-query count-query]}
+                                    (query->sql query entity version query-options context)
+                                    st (when count-query
+                                         {:count (jdbc/get-result-count count-query)})
+                                    add-sleep #(format "select pg_sleep(%f), * from (%s) as placeholder_name"
+                                                       (float diagnostic-inter-row-sleep) %)
+                                    results-query (cond-> results-query
+                                                    diagnostic-inter-row-sleep (update 0 add-sleep))]
+                                (jdbc/update-local-timeouts query-deadline-ns 1)
+                                (jdbc/call-with-array-converted-query-rows
+                                 results-query
+                                 (when diagnostic-inter-row-sleep {:fetch-size 1})
+                                 #(stream-rows % out st))))
+                            (finally
+                              (when query-monitor-id
+                                (qmon/forget query-monitor query-monitor-id)))))
+                        (catch SQLException ex
+                          (if (jdbc/local-timeout-ex? ex)
+                            (throw-timeout)
+                            (throw ex)))
+                        (catch ExceptionInfo ex
+                          (cond
+                            ;; If we've already sent the response to
+                            ;; the client (i.e. status has been
+                            ;; delivered) then send something "useful"
+                            ;; to the client (at the end of the
+                            ;; truncated JSON) before we close the
+                            ;; connection.
 
-(defn preferred-produce-streaming-body
-  [version query-map context]
-  (let [{:keys [scf-read-db url-prefix warn-experimental log-queries query-id]
-         :or {warn-experimental true}} context
-        query-config (select-keys context [:node-purge-ttl :add-agent-report-filter])
-        {:keys [query remaining-query entity query-options]}
-        (user-query->engine-query version query-map query-config warn-experimental)]
+                            (and (realized? status)
+                                 (= :puppetlabs.puppetdb.query/timeout (:kind (ex-data ex))))
+                            (let [msg (.getMessage ex)]
+                              (log/warn msg)
+                              (.write out msg) (.flush out))
 
-    (when log-queries
-      ;; Log origin and AST of incoming query
-      (let [{:keys [origin query]} query-map]
-        (log/infof "PDBQuery:%s:%s"
-                   query-id (-> (sorted-map :origin origin :ast query)
-                                json/generate-string))))
+                            (and (realized? status) (jdbc/clj-jdbc-termination-ex? ex))
+                            (let [msg (query-terminated-msg query-id origin)]
+                              (log/warn msg)
+                              (.write out msg) (.flush out))
 
-    (try
-      (let [munge-fn (get-munge-fn entity version query-options url-prefix)
-            stream-ctx (select-keys context [:log-queries :pretty-print :query-id])
-            {:keys [status stream]} (body-stream scf-read-db
-                                                 (coerce-from-json remaining-query)
-                                                 entity version query-options
-                                                 munge-fn
-                                                 stream-ctx)
-            {:keys [count error]} @status]
-        (when error
-          (throw error))
-        (cond-> (http/json-response* stream)
-          count (http/add-headers {:count count})))
-      (catch JsonParseException ex
-        (log/error ex (trs "Unparsable query: {0} {1} {2}" query-id query query-options))
-        (http/error-response ex))
-      (catch IllegalArgumentException ex ;; thrown by (at least) munge-fn
-        (log/error ex (trs "Invalid query: {0} {1} {2}" query-id query query-options))
-        (http/error-response ex))
-      (catch PSQLException ex
-        (when-not (= (.getSQLState ex) (jdbc/sql-state :invalid-regular-expression))
-          (throw ex))
-        (do
-          (log/debug ex (trs "Invalid query regex: {0} {1} {2}" query-id query query-options))
-          (http/error-response ex))))))
+                            :else (throw ex))))))
+
+                  ;; These delivers may not do anything, i.e. if the
+                  ;; query has already started.
+                  (catch Throwable ex
+                    (cond
+                      (identical? quiet-exit ex)
+                      (when-not (realized? status)
+                        (log/error ex (trs "Impossible situation: query streamer exiting without delivery")))
+
+                      ;; connection to client has been closed at this point
+                      (realized? status)
+                      (cond
+                        (and (instance? ExceptionInfo ex)
+                             (= :puppetlabs.puppetdb.query/timeout (:kind (ex-data ex))))
+                        (log/warn (.getMessage ex))
+
+                        (and (instance? ExceptionInfo ex)
+                             (jdbc/clj-jdbc-termination-ex? ex))
+                        (log/warn (query-terminated-msg query-id origin))
+
+                        :else
+                        (let [msg (trs "Query streaming failed: {0} {1}" query query-options)]
+                          (log/error ex msg)
+                          (throw ex)))
+
+                      :else (deliver status {:error ex}))))))}))
 
 ;; for testing via with-redefs
 (def munge-fn-hook identity)
 
-(defn- deprecated-produce-streaming-body
-  [version query-map context]
-  (let [{:keys [scf-read-db url-prefix warn-experimental pretty-print
-                log-queries query-id]
-         :or {warn-experimental true}} context
-        query-config (select-keys context [:node-purge-ttl :add-agent-report-filter])
-        {:keys [query remaining-query entity query-options]}
-        (user-query->engine-query version query-map query-config warn-experimental)
-        origin (:origin query-map)]
-
-    (when log-queries
-      ;; Log origin and AST of incoming query
-      (let [query (:query query-map)]
-        (log/infof "PDBQuery:%s:%s"
-                   query-id (-> (sorted-map :origin origin :ast query)
-                                json/generate-string))))
-
-    (try
-      (jdbc/with-transacted-connection scf-read-db
-        (let [munge-fn (get-munge-fn entity version query-options url-prefix)
-              {:keys [results-query count-query]}
-              (-> remaining-query
-                  coerce-from-json
-                  (query->sql entity version query-options
-                              (select-keys context [:log-queries :query-id])))
-              query-error (promise)
-              resp (http/streamed-response
-                    buffer
-                    (with-log-mdc ["pdb-query-id" query-id
-                                   "pdb-query-origin" origin]
-                      (try (jdbc/with-transacted-connection scf-read-db
-                             (jdbc/call-with-array-converted-query-rows
-                              results-query
-                              (comp #(http/stream-json % buffer pretty-print)
-                                    #(do (when-not (instance? PGobject %)
-                                           (first %))
-                                         (deliver query-error nil) %)
-                                    (munge-fn-hook munge-fn))))
-                           ;; catch throwable to make sure any trouble is forwarded to the
-                           ;; query-error promise below. If something throws and is not passed
-                           ;; along the deref will block indefinitely.
-                           (catch Throwable e
-                             (deliver query-error e)))))]
-          (if @query-error
-            (throw @query-error)
-            (cond-> (http/json-response* resp)
-                    count-query (http/add-headers {:count (jdbc/get-result-count count-query)})))))
-      (catch com.fasterxml.jackson.core.JsonParseException e
-        (log/error
-         e
-         (trs "Error executing query ''{0}'' with query options ''{1}''. Returning a 400 error code."
-              query query-options))
-        (http/error-response e))
-      (catch IllegalArgumentException e
-        (log/error
-         e
-         (trs
-          "Error executing query ''{0}'' with query options ''{1}''. Returning a 400 error code."
-          query query-options))
-        (http/error-response e))
-      (catch org.postgresql.util.PSQLException e
-        (if (= (.getSQLState e) (jdbc/sql-state :invalid-regular-expression))
-          (do (log/debug e (trs "Caught PSQL processing exception"))
-              (http/error-response (.getMessage e)))
-          (throw e))))))
-
 (pls/defn-validated produce-streaming-body
   "Given a query, and database connection, return a Ring response with
    the query results. query-map is a clojure map of the form
@@ -495,9 +516,48 @@
   (let [context (assoc context :query-id query-uuid)]
     (with-log-mdc ["pdb-query-id" query-uuid
                    "pdb-query-origin" (:origin query-map)]
-      (if use-preferred-streaming-method?
-        (preferred-produce-streaming-body version query-map context)
-        (deprecated-produce-streaming-body version query-map context)))))
+      (let [{:keys [scf-read-db url-prefix warn-experimental log-queries query-id]
+             :or {warn-experimental true}} context
+            query-config (select-keys context [:node-purge-ttl :add-agent-report-filter])
+            {:keys [query remaining-query entity query-options]}
+            (user-query->engine-query version query-map query-config warn-experimental)]
+
+        (when log-queries
+          ;; Log origin and AST of incoming query
+          (let [{:keys [origin query]} query-map]
+            (log/infof "PDBQuery:%s:%s"
+                       query-id (-> (sorted-map :origin origin :ast query)
+                                    json/generate-string))))
+
+        (try
+          (let [munge-fn (get-munge-fn entity version query-options url-prefix)
+                stream-ctx (select-keys context [:log-queries :pretty-print :query-id
+                                                 :query-deadline-ns
+                                                 :query-monitor
+                                                 :query-monitor-id])
+                {:keys [status stream]} (body-stream scf-read-db
+                                                     (coerce-from-json remaining-query)
+                                                     entity version query-options
+                                                     (munge-fn-hook munge-fn)
+                                                     stream-ctx)
+                {:keys [count error]} @status]
+            (when error
+              (throw error))
+            (cond-> (-> (http/json-response* stream)
+                        (assoc :puppetlabs.puppetdb.query/streaming true))
+              count (http/add-headers {:count count})))
+          (catch JsonParseException ex
+            (log/error ex (trs "Unparsable query: {0} {1} {2}" query-id query query-options))
+            (http/error-response ex))
+          (catch IllegalArgumentException ex ;; thrown by (at least) munge-fn
+            (log/error ex (trs "Invalid query: {0} {1} {2}" query-id query query-options))
+            (http/error-response ex))
+          (catch PSQLException ex
+            (when-not (= (.getSQLState ex) (jdbc/sql-state :invalid-regular-expression))
+              (throw ex))
+            (do
+              (log/debug ex (trs "Invalid query regex: {0} {1} {2}" query-id query query-options))
+              (http/error-response ex))))))))
 
 (pls/defn-validated object-exists? :- s/Bool
   "Returns true if an object exists."
