@@ -4,7 +4,6 @@
             [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.scf.migrate :as migrate
              :refer [applied-migrations
-                     create-indexes
                      desired-schema-version
                      initialize-schema
                      migrations
@@ -25,6 +24,8 @@
             [puppetlabs.puppetdb.scf.partitioning :as part]
             [clojure.string :as str])
   (:import (java.time ZoneId ZonedDateTime)
+           (java.time.format DateTimeFormatter)
+           (java.time.temporal ChronoUnit)
            (java.sql Timestamp)))
 
 (use-fixtures :each tdb/call-with-test-db)
@@ -533,6 +534,7 @@
                      "group by gid"))))))))
 
 (deftest migration-50-remove-historical-catalogs
+  ;; Completely retired/obsolete
   (jdbc/with-db-connection *db*
     (clear-db-for-testing!)
     (fast-forward-to-migration! 49)
@@ -1288,18 +1290,6 @@
                (format "SELECT reloptions FROM pg_class WHERE relname = '%s' AND CAST(reloptions as text) LIKE '{autovacuum_vacuum_scale_factor=%s}'"
                        table factor)))))))
 
-(deftest autovacuum-analyze-scale-factor-test
-  (clear-db-for-testing!)
-  ;; intentionally apply all of the migrations before looking to ensure our autovacuum scale factors aren't lost
-  (fast-forward-to-migration! (desired-schema-version))
-  (create-indexes)
-  (let [values {"catalog_resources" "0.01"}]
-    (doseq [[table factor] values]
-      (is (= [{:reloptions [(format "autovacuum_analyze_scale_factor=%s" factor)]}]
-             (jdbc/query-to-vec
-               (format "SELECT reloptions FROM pg_class WHERE relname = '%s' AND CAST(reloptions as text) LIKE '{autovacuum_analyze_scale_factor=%s}'"
-                       table factor)))))))
-
 (deftest migration-74-changes-hash-of-reports
   (let [current-time (to-timestamp (now))
         old-hash-fn (fn [{:keys [certname puppet_version report_format configuration_version
@@ -1705,3 +1695,514 @@
                    :deferrable? "NO"},
                   :same nil}]}
                (diff-schema-maps before-migration (schema-info-map *db*))))))))
+
+;; Helper functions for migration-81-resource-events-declarative-partitioning
+;; and migration-82-reports-declarative-partitioning
+(defn diff-date-suffix
+  [date]
+  (let [formatter (.withZone (DateTimeFormatter/BASIC_ISO_DATE) (ZoneId/of "UTC"))]
+    (str/lower-case
+      (.format date formatter))))
+
+(defn resource-events-partition-day-index-diff-template
+  "Generates the expected primary key indix for one resource-event's partition
+  day table."
+  [date]
+  (let [date-suffix (diff-date-suffix date)]
+    [{:left-only nil,
+       :right-only
+       {:schema "public",
+        :table (format "resource_events_%s" date-suffix),
+        :index (format "resource_events_%s_pkey" date-suffix),
+        :index_keys ["event_hash" "\"timestamp\""],
+        :type "btree",
+        :unique? true,
+        :functional? false,
+        :is_partial false,
+        :primary? true,
+        :user "pdb_test"},
+       :same nil}]))
+
+(defn get-formatted-start-of-day
+  "Returns a formatted date like '2023-01-13 00:00:00+00' as seen in Postgres
+  constraints retrieved from the schema. This assumes that Postgres is running
+  in the systemDefault timezone"
+  [date]
+  (let [date-formatter (.withZone (DateTimeFormatter/ofPattern "yyyy-MM-dd HH:mm:ssx") (ZoneId/systemDefault))]
+    (-> date
+        (.truncatedTo (ChronoUnit/DAYS)) ;; this is a ZonedDateTime
+        (.format date-formatter))))
+
+(defn resource-events-partition-day-constraint-diff-template
+  "generates the expected diff of constraints for one resource-event's
+  partition day table"
+  [date]
+  (let [date-suffix (diff-date-suffix date)
+        formatted-start-of-day (get-formatted-start-of-day date)
+        formatted-start-of-next-day (get-formatted-start-of-day (.plusDays date 1))]
+    [{:left-only
+      {:constraint_name
+       (format "(((\"timestamp\" >= '%s'::timestamp with time zone) AND (\"timestamp\" < '%s'::timestamp with time zone)))" formatted-start-of-day formatted-start-of-next-day),
+       :table_name (format "resource_events_%s" date-suffix),
+       :constraint_type "CHECK",
+       :initially_deferred "NO",
+       :deferrable? "NO"},
+      :right-only nil,
+      :same nil}
+     {:left-only nil,
+      :right-only
+      {:constraint_name (format "resource_events_%s_pkey" date-suffix),
+       :table_name (format "resource_events_%s" date-suffix),
+       :constraint_type "PRIMARY KEY",
+       :initially_deferred "NO",
+       :deferrable? "NO"},
+      :same nil}]))
+
+(defn generate-diff-sequence
+  "Stamp out a set of diffs from the given template"
+  [template-fn]
+  (let [now (ZonedDateTime/now)
+        days (range -4 4)]
+    (flatten
+      (map (fn [day-offset]
+             (template-fn (part/to-zoned-date-time (.plusDays now day-offset))))
+           days))))
+
+(deftest migration-81-resource-events-declarative-partitioning
+  (testing "resource_events table declarative partitioning migration"
+    (jdbc/with-db-connection *db*
+      (clear-db-for-testing!)
+      (fast-forward-to-migration! 80)
+      (let [before-migration (schema-info-map *db*)
+            exp-resource-events-indices-diff
+              [{:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_timestamp_idx",
+                 :index_keys ["\"timestamp\""],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_containing_class_idx",
+                 :index_keys ["containing_class"],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_pkey",
+                 :index_keys ["event_hash"],
+                 :type "btree",
+                 :unique? true,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? true,
+                 :user "pdb_test"},
+                :right-only nil,
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_pkey",
+                 :index_keys ["event_hash", "\"timestamp\""],
+                 :type "btree",
+                 :unique? true,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? true,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_property_idx",
+                 :index_keys ["property"],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_reports_id_idx",
+                 :index_keys ["report_id"],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_resource_title_idx",
+                 :index_keys ["resource_title"],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_status_for_corrective_change_idx",
+                 :index_keys ["status"],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial true,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "resource_events",
+                 :index "resource_events_resource_timestamp",
+                 :index_keys ["resource_type" "resource_title" "\"timestamp\""],
+                 :type "btree",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}]
+            exp-idx-diff (concat exp-resource-events-indices-diff
+                                 (generate-diff-sequence resource-events-partition-day-index-diff-template))
+            exp-constraint-diff (generate-diff-sequence resource-events-partition-day-constraint-diff-template)
+            expected-diff
+              {:index-diff (set exp-idx-diff),
+               :table-diff nil,
+               :constraint-diff (set exp-constraint-diff)}]
+        (apply-migration-for-testing! 81)
+        (let [raw-diff (diff-schema-maps before-migration (schema-info-map *db*))
+              diff (-> raw-diff
+                     (update :index-diff set)
+                     (update :constraint-diff set))]
+          (is (= expected-diff diff)))))))
+
+(defn report-partition-day-index-diff-template
+  "Generates the expected diff of primary key and producer timestamp indices
+  for one report's partition day table."
+  [date]
+  (let [date-suffix (diff-date-suffix date)
+        common-partition-indexes
+          [{:left-only nil,
+             :right-only
+             {:schema "public",
+              :table (format "reports_%s" date-suffix),
+              :index (format "reports_%s_pkey" date-suffix),
+              :index_keys ["id" "producer_timestamp"],
+              :type "btree",
+              :unique? true,
+              :functional? false,
+              :is_partial false,
+              :primary? true,
+              :user "pdb_test"},
+             :same nil}
+             {:left-only nil,
+              :right-only
+              {:schema "public",
+               :table (format "reports_%s" date-suffix),
+               :index (format "reports_%s_encode_producer_timestamp_idx" date-suffix),
+               :index_keys ["encode(hash, 'hex'::text)" "producer_timestamp" ],
+               :type "btree",
+               :unique? true,
+               :functional? true,
+               :is_partial false,
+               :primary? false,
+               :user "pdb_test"},
+              :same nil}]
+        ;; Postgresql 11.18 fixed a bug in naming of partitioned indexes that might
+        ;; be contributing to the changed naming of reports_noop_idx in partitions in
+        ;; PG 11.17 testing...
+        ;; https://www.postgresql.org/docs/11/release-11-18.html#id-1.11.6.5.4
+        pg11-18-db [11 18]
+        {current-db-version :version} (sutils/db-metadata)]
+        (if (neg? (compare current-db-version pg11-18-db))
+          ;; In PG < 11.18 postgresql mysteriously has a different naming for
+          ;; the automatically generated noop indexes on the partition
+          ;; tables...
+          (conj common-partition-indexes
+                {:left-only
+                 {:index (format "reports_noop_idx_%s" date-suffix)},
+                  :right-only
+                  {:index (format "reports_%s_noop_idx" date-suffix)},
+                   :same {:schema "public",
+                   :table (format "reports_%s" date-suffix),
+                   :index_keys ["noop"],
+                   :type "btree",
+                   :unique? false,
+                   :functional? false,
+                   :is_partial true,
+                   :primary? false,
+                   :user "pdb_test"}})
+          common-partition-indexes)))
+
+(defn report-partition-day-constraint-diff-template
+  "Generates the expected diff of a constraints for one report's partition day table"
+  [date]
+  (let [date-suffix (diff-date-suffix date)
+        formatted-start-of-day (get-formatted-start-of-day date)
+        formatted-start-of-next-day (get-formatted-start-of-day (.plusDays date 1))
+        table-name (format "reports_%s" date-suffix)]
+    [{:left-only
+      {:constraint_name
+       (format "(((producer_timestamp >= '%s'::timestamp with time zone) AND (producer_timestamp < '%s'::timestamp with time zone)))" formatted-start-of-day formatted-start-of-next-day)
+       :table_name (format "reports_%s" date-suffix)
+       :constraint_type "CHECK"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :right-only nil
+      :same nil}
+     {:left-only nil
+      :right-only
+      {:constraint_name (format "reports_%s_pkey" date-suffix)
+       :table_name (format "reports_%s" date-suffix)
+       :constraint_type "PRIMARY KEY"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :same nil}
+     {:left-only
+      {:constraint_name (format "reports_prod_fkey_%s" date-suffix)
+       :table_name table-name
+       :constraint_type "FOREIGN KEY"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :right-only nil
+      :same nil}
+     {:left-only
+      {:constraint_name (format "reports_certname_fkey_%s" date-suffix)
+       :table_name table-name
+       :constraint_type "FOREIGN KEY"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :right-only nil
+      :same nil}
+     {:left-only
+      {:constraint_name (format "reports_status_fkey_%s" date-suffix)
+       :table_name table-name
+       :constraint_type "FOREIGN KEY"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :right-only nil
+      :same nil}
+     {:left-only
+      {:constraint_name (format "reports_env_fkey_%s" date-suffix)
+       :table_name table-name
+       :constraint_type "FOREIGN KEY"
+       :initially_deferred "NO"
+       :deferrable? "NO"}
+      :right-only nil
+      :same nil}]))
+
+(deftest migration-82-reports-declarative-partitioning
+  (testing "reports table declarative partitioning migration"
+    (jdbc/with-db-connection *db*
+      (clear-db-for-testing!)
+      (fast-forward-to-migration! 81)
+      (let [before-migration (schema-info-map *db*)
+            exp-reports-indices-diff
+              [{:left-only
+                {:schema "public"
+                 :table "reports"
+                 :index "reports_hash_expr_idx"
+                 :index_keys ["encode(hash, 'hex'::text)"]
+                 :type "btree"
+                 :unique? true
+                 :functional? true
+                 :is_partial false
+                 :primary? false
+                 :user "pdb_test"}
+                :right-only nil
+                :same nil}
+               {:left-only nil
+                :right-only
+                {:schema "public"
+                 :table "reports"
+                 :index "reports_hash_expr_idx"
+                 :index_keys ["encode(hash, 'hex'::text)" "producer_timestamp"]
+                 :type "btree"
+                 :unique? true
+                 :functional? true
+                 :is_partial false
+                 :primary? false
+                 :user "pdb_test"}
+                :same nil}
+               {:left-only
+                {:schema "public"
+                 :table "reports"
+                 :index "reports_pkey"
+                 :index_keys ["id"]
+                 :type "btree"
+                 :unique? true
+                 :functional? false
+                 :is_partial false
+                 :primary? true
+                 :user "pdb_test"}
+                :right-only nil
+                :same nil}
+               {:left-only nil
+                :right-only
+                {:schema "public"
+                 :table "reports"
+                 :index "reports_pkey"
+                 :index_keys ["id" "producer_timestamp"]
+                 :type "btree"
+                 :unique? true
+                 :functional? false
+                 :is_partial false
+                 :primary? true
+                 :user "pdb_test"}
+                :same nil}
+               {:left-only nil
+                :right-only
+                {:schema "public"
+                 :table "reports"
+                 :index "idx_reports_compound_id"
+                 :index_keys ["producer_timestamp" "certname" "hash"]
+                 :type "btree"
+                 :unique? false
+                 :functional? false
+                 :is_partial true
+                 :primary? false
+                 :user "pdb_test"}
+                :same nil}]
+
+              exp-reports-constraint-diff
+              [{:left-only
+                {:constraint_name "reports_prod_fkey",
+                 :table_name "reports",
+                 :constraint_type "FOREIGN KEY",
+                 :initially_deferred "NO",
+                 :deferrable? "NO"},
+                :right-only nil,
+                :same nil}
+               {:left-only
+                {:constraint_name "reports_certname_fkey",
+                 :table_name "reports",
+                 :constraint_type "FOREIGN KEY",
+                 :initially_deferred "NO",
+                 :deferrable? "NO"},
+                :right-only nil,
+                :same nil}
+               {:left-only
+                {:constraint_name "reports_status_fkey",
+                 :table_name "reports",
+                 :constraint_type "FOREIGN KEY",
+                 :initially_deferred "NO",
+                 :deferrable? "NO"},
+                :right-only nil,
+                :same nil}
+               {:left-only
+                {:constraint_name "reports_env_fkey",
+                 :table_name "reports",
+                 :constraint_type "FOREIGN KEY",
+                 :initially_deferred "NO",
+                 :deferrable? "NO"},
+                :right-only nil,
+                :same nil}]
+              exp-idx-diff (concat exp-reports-indices-diff
+                                   (generate-diff-sequence report-partition-day-index-diff-template))
+              exp-constraint-diff (concat exp-reports-constraint-diff
+                                          (generate-diff-sequence report-partition-day-constraint-diff-template))
+            expected-diff
+              {:index-diff (set exp-idx-diff)
+               :table-diff nil
+               :constraint-diff (set exp-constraint-diff)}]
+        (apply-migration-for-testing! 82)
+        (let [raw-diff (diff-schema-maps before-migration (schema-info-map *db*))
+              diff (-> raw-diff
+                     (update :index-diff set)
+                     (update :constraint-diff set))]
+          (is (= expected-diff diff)))))))
+
+(deftest migration-83-creates-trigram-indexes
+  (jdbc/with-db-connection *db*
+    (clear-db-for-testing!)
+    (fast-forward-to-migration! 82)
+    (let [before-migration (schema-info-map *db*)]
+      (apply-migration-for-testing! 83)
+      (is (= {:index-diff
+              [{:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "catalog_resources",
+                 :index "catalog_resources_file_trgm",
+                 :index_keys ["file"],
+                 :type "gin",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial true,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}
+               {:left-only nil,
+                :right-only
+                {:schema "public",
+                 :table "fact_paths",
+                 :index "fact_paths_path_trgm",
+                 :index_keys ["path"],
+                 :type "gist",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial false,
+                 :primary? false,
+                 :user "pdb_test"},
+                :same nil}],
+              :table-diff nil,
+              :constraint-diff nil}
+             (diff-schema-maps before-migration (schema-info-map *db*)))))))
+
+(deftest migration-84-removes-catalog-resources-file-trgm-index
+  (jdbc/with-db-connection *db*
+    (clear-db-for-testing!)
+    (fast-forward-to-migration! 83)
+    (let [before-migration (schema-info-map *db*)]
+      (apply-migration-for-testing! 84)
+      (is (= {:index-diff
+              [{:left-only
+                {:schema "public",
+                 :table "catalog_resources",
+                 :index "catalog_resources_file_trgm",
+                 :index_keys ["file"],
+                 :type "gin",
+                 :unique? false,
+                 :functional? false,
+                 :is_partial true,
+                 :primary? false,
+                 :user "pdb_test"},
+                :right-only nil,
+                :same nil}],
+              :table-diff nil,
+              :constraint-diff nil}
+             (diff-schema-maps before-migration (schema-info-map *db*)))))
+    (testing "Check that autovacuum_analyze_scale_factor is removed from catalog_resources"
+      (is (= []
+             (jdbc/query-to-vec
+               "SELECT reloptions FROM pg_class WHERE relname = 'catalog_resources' AND CAST(reloptions as text) LIKE '%autovacuum_analyze_scale_factor%'"))))))
